"""
Gaussian distributions for pygama
"""

import numba as nb
import numpy as np

from pygama.math.functions.error_function import nb_erf
from pygama.math.functions.pygama_continuous import PygamaContinuous
from pygama.utils import numba_math_defaults as nb_defaults


@nb.njit(**nb_defaults(parallel=False))
def nb_gauss(x: np.ndarray, mu: float, sigma: float) -> np.ndarray:
    """
    Gaussian, unnormalised for use in building PDFs
    As a Numba JIT function, it runs slightly faster than
    'out of the box' functions.

    Parameters
    ----------
    x
        Input data
    mu
        The centroid of the Gaussian
    sigma
        The standard deviation of the Gaussian

    .. note::
        TODO:: remove this in favor of using nb_gauss_pdf with a different normalization
    """

    if sigma == 0:
        invs = np.inf
    else:
        invs = 1.0 / sigma
    z = (x - mu) * invs
    return np.exp(-0.5 * z**2)


@nb.njit(**nb_defaults(parallel=False))
def nb_gauss_amp(x: np.ndarray, mu: float, sigma: float, a: float) -> np.ndarray:
    """
    Gaussian with height as a parameter for FWHM etc.
    As a Numba JIT function, it runs slightly faster than
    'out of the box' functions.

    Parameters
    ----------
    x
        Input data
    mu
        The centroid of the Gaussian
    sigma
        The standard deviation of the Gaussian
    a
        The amplitude of the Gaussian

    .. note::
        TODO:: potentially remove this, redundant with ``nb_gauss_scaled_pdf``
    """

    if sigma == 0:
        invs = np.inf
    else:
        invs = 1.0 / sigma
    z = (x - mu) * invs
    return a * np.exp(-0.5 * z**2)


@nb.njit(**nb_defaults(parallel=False))
def nb_gauss_pdf(x: np.ndarray, mu: float, sigma: float) -> np.ndarray:
    r"""
    Normalised Gaussian PDF, w/ args: mu, sigma. The support is :math:`(-\infty, \infty)`

    .. math::
        pdf(x, \mu, \sigma) = \frac{1}{\sqrt{2\pi}}e^{(\frac{x-\mu}{\sigma}^2)/2}

    As a Numba JIT function, it runs slightly faster than
    'out of the box' functions.

    Parameters
    ----------
    x
        The input data
    mu
        The centroid of the Gaussian
    sigma
        The standard deviation of the Gaussian
    """

    if sigma == 0:
        invs = np.inf
    else:
        invs = 1.0 / sigma
    z = (x - mu) * invs
    invnorm = invs / np.sqrt(2 * np.pi)
    return np.exp(-0.5 * z**2) * invnorm


@nb.njit(**nb_defaults(parallel=False))
def nb_gauss_cdf(x: np.ndarray, mu: float, sigma: float) -> np.ndarray:
    r"""
    Gaussian CDF, w/ args: mu, sigma. The support is :math:`(-\infty, \infty)`

    .. math::
        cdf(x, \mu,\sigma) =  \frac{1}{2}\left[1+\text{erf}(\frac{x-\mu}{\sigma\sqrt{2}})\right]

    As a Numba JIT function, it runs slightly faster than
    'out of the box' functions.

    Parameters
    ----------
    x
        The input data
    mu
        The centroid of the Gaussian
    sigma
        The standard deviation of the Gaussian
    """

    if sigma == 0:
        invs = np.inf
    else:
        invs = 1.0 / sigma

    return 1 / 2 * (1 + nb_erf(invs * (x - mu) / (np.sqrt(2))))


@nb.njit(**nb_defaults(parallel=False))
def nb_gauss_scaled_pdf(
    x: np.ndarray, area: float, mu: float, sigma: float
) -> np.ndarray:
    """
    Gaussian with height as a parameter for fwhm etc.
    As a Numba JIT function, it runs slightly faster than
    'out of the box' functions.

    Parameters
    ----------
    x
        Input data
    area
        The number of counts in the signal
    mu
        The centroid of the Gaussian
    sigma
        The standard deviation of the Gaussian
    """

    return area * nb_gauss_pdf(x, mu, sigma)


@nb.njit(**nb_defaults(parallel=False))
def nb_gauss_scaled_cdf(
    x: np.ndarray, area: float, mu: float, sigma: float
) -> np.ndarray:
    """
    Gaussian CDF scaled by the number of signal counts for extended binned fits
    As a Numba JIT function, it runs slightly faster than
    'out of the box' functions.

    Parameters
    ----------
    x
        Input data
    area
        The number of counts in the signal
    mu
        The centroid of the Gaussian
    sigma
        The standard deviation of the Gaussian
    """

    return area * nb_gauss_cdf(x, mu, sigma)


class GaussianGen(PygamaContinuous):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.x_lo = -1 * np.inf
        self.x_hi = np.inf

    def _pdf(self, x: np.ndarray) -> np.ndarray:
        x.flags.writeable = True
        return nb_gauss_pdf(x, 0, 1)

    def _cdf(self, x: np.ndarray) -> np.ndarray:
        return nb_gauss_cdf(x, 0, 1)

    def get_pdf(self, x: np.ndarray, mu: float, sigma: float) -> np.ndarray:
        return nb_gauss_pdf(x, mu, sigma)

    def get_cdf(self, x: np.ndarray, mu: float, sigma: float) -> np.ndarray:
        return nb_gauss_cdf(x, mu, sigma)

    def pdf_norm(
        self, x: np.ndarray, x_lo: float, x_hi: float, mu: float, sigma: float
    ) -> np.ndarray:
        return self._pdf_norm(x, x_lo, x_hi, mu, sigma)

    def cdf_norm(
        self, x: np.ndarray, x_lo: float, x_hi: float, mu: float, sigma: float
    ) -> np.ndarray:
        return self._cdf_norm(x, x_lo, x_hi, mu, sigma)

    def pdf_ext(
        self,
        x: np.ndarray,
        x_lo: float,
        x_hi: float,
        area: float,
        mu: float,
        sigma: float,
    ) -> np.ndarray:
        return np.diff(nb_gauss_scaled_cdf(np.array([x_lo, x_hi]), area, mu, sigma))[
            0
        ], nb_gauss_scaled_pdf(x, area, mu, sigma)

    def cdf_ext(
        self, x: np.ndarray, area: float, mu: float, sigma: float
    ) -> np.ndarray:
        return nb_gauss_scaled_cdf(x, area, mu, sigma)

    def required_args(self) -> tuple[str, str]:
        return "mu", "sigma"


gaussian = GaussianGen(name="gaussian")



"""
pygama convenience functions for linearly fitting data
"""

from typing import Optional, Union

import numpy as np


def linear_fit_by_sums(
    x: np.ndarray, y: np.ndarray, var: Optional[Union[float, np.ndarray]] = 1
) -> tuple[float, float]:
    """
    Fast computation of weighted linear least squares fit to a linear model

    Note: doesn't compute covariances. If you want covariances, just use polyfit

    Parameters
    ----------
    x
        x values for the fit
    y
        y values for the fit
    var
        The variances for each y-value

    Returns
    -------
    (m, b)
        The slope (m) and y-intercept (b) of the best fit (in the least-squares
        sense) of the data to y = mx + b
    """
    y = y / var
    x = x / var
    sum_wts = len(y) / var if np.isscalar(var) else np.sum(1 / var)
    sum_x = np.sum(x)
    sum_xx = np.sum(x * x)
    sum_y = np.sum(y)
    sum_yx = np.sum(y * x)
    m = (sum_wts * sum_yx - sum_y * sum_x) / (sum_wts * sum_xx - sum_x**2)
    b = (sum_y - m * sum_x) / sum_wts
    return m, b


def fit_simple_scaling(
    x: np.ndarray, y: np.ndarray, var: Optional[Union[float, np.ndarray]] = 1
) -> tuple[float, float]:
    """
    Fast computation of weighted linear least squares fit to a simple scaling

    I.e. y = scale * x. Returns the best fit scale parameter and its variance.

    Parameters
    ----------
    x
        x values for the fit
    y
        y values for the fit
    var
        The variances for each y-value

    Returns
    -------
    scale, scale_var
        The scale parameter and its variance
    """
    x = np.asarray(x)
    y = np.asarray(y)
    scale_var = 1 / np.sum(x * x / var)
    scale = np.sum(y * x / var) * scale_var
    return scale, scale_var



import json
import os
from pathlib import Path

import awkward as ak
import numpy as np
import pytest
from lgdo import lh5

from pygama.hit import build_hit
from pygama.hit.build_hit import _reorder_table_operations

config_dir = Path(__file__).parent / "configs"


def test_ops_reorder():
    assert list(_reorder_table_operations({}).keys()) == []

    ops = {
        "out1": {"expression": "out2 + out3 * outy"},
        "out2": {"expression": "log(out4)"},
        "out3": {"expression": "outx + 2"},
        "out4": {"expression": "outz + out3"},
    }
    assert list(_reorder_table_operations(ops).keys()) == [
        "out3",
        "out4",
        "out2",
        "out1",
    ]


def test_basics(dsp_test_file, tmp_dir):
    outfile = f"{tmp_dir}/LDQTA_r117_20200110T105115Z_cal_geds_hit.lh5"

    build_hit(
        dsp_test_file,
        outfile=outfile,
        hit_config=f"{config_dir}/basic-hit-config.json",
        wo_mode="overwrite",
    )

    assert os.path.exists(outfile)
    assert lh5.ls(outfile, "/geds/") == ["geds/hit"]

    tbl = lh5.read("geds/hit", outfile)
    assert tbl.calE.attrs == {"datatype": "array<1>{real}", "units": "keV"}


def test_illegal_arguments(dsp_test_file):
    with pytest.raises(ValueError):
        build_hit(dsp_test_file)

    with pytest.raises(ValueError):
        build_hit(
            dsp_test_file,
            hit_config=f"{config_dir}/basic-hit-config.json",
            lh5_tables_config={},
        )

    with pytest.raises(ValueError):
        build_hit(
            dsp_test_file,
            lh5_tables=[],
            lh5_tables_config={},
        )


def test_lh5_table_configs(dsp_test_file, tmp_dir):
    outfile = f"{tmp_dir}/LDQTA_r117_20200110T105115Z_cal_geds_hit.lh5"

    lh5_tables_config = {"/geds/dsp": f"{config_dir}/basic-hit-config.json"}

    build_hit(
        dsp_test_file,
        outfile=outfile,
        lh5_tables_config=lh5_tables_config,
        wo_mode="overwrite",
    )

    assert os.path.exists(outfile)
    assert lh5.ls(outfile, "/geds/") == ["geds/hit"]

    lh5_tables_config = {
        "/geds/dsp": {
            "outputs": ["calE", "AoE"],
            "operations": {
                "calE": {
                    "expression": "sqrt(a + b * trapEmax**2)",
                    "parameters": {"a": 1.23, "b": 42.69},
                },
                "AoE": {"expression": "A_max/calE"},
            },
        }
    }

    build_hit(
        dsp_test_file,
        outfile=outfile,
        lh5_tables_config=lh5_tables_config,
        wo_mode="overwrite",
    )

    assert os.path.exists(outfile)
    assert lh5.ls(outfile, "/geds/") == ["geds/hit"]


def test_outputs_specification(dsp_test_file, tmp_dir):
    outfile = f"{tmp_dir}/LDQTA_r117_20200110T105115Z_cal_geds_hit.lh5"

    build_hit(
        dsp_test_file,
        outfile=outfile,
        hit_config=f"{config_dir}/basic-hit-config.json",
        wo_mode="overwrite",
    )

    obj = lh5.read("/geds/hit", outfile)
    assert sorted(obj.keys()) == ["A_max", "AoE", "calE"]


def test_aggregation_outputs(dsp_test_file, tmp_dir):
    outfile = f"{tmp_dir}/LDQTA_r117_20200110T105115Z_cal_geds_hit.lh5"

    build_hit(
        dsp_test_file,
        outfile=outfile,
        hit_config=f"{config_dir}/aggregations-hit-config.json",
        wo_mode="overwrite",
    )

    obj = lh5.read("/geds/hit", outfile)
    assert sorted(obj.keys()) == [
        "aggr1",
        "aggr2",
        "is_valid_rt",
        "is_valid_t0",
        "is_valid_tmax",
    ]

    df = lh5.read_as("geds/hit", outfile, "pd")

    # aggr1 consists of 3 bits --> max number can be 7, aggr2 consists of 2 bits so max number can be 3
    assert not (df["aggr1"] > 7).any()
    assert not (df["aggr2"] > 3).any()

    def get_bit(x, n):
        """bit numbering from right to left, starting with bit 0"""
        return x & (1 << n) != 0

    df["bit0_check"] = df.apply(lambda row: get_bit(row["aggr1"], 0), axis=1)
    are_identical = df["bit0_check"].equals(df.is_valid_rt)
    assert are_identical

    df["bit1_check"] = df.apply(lambda row: get_bit(row["aggr1"], 1), axis=1)
    are_identical = df["bit1_check"].equals(df.is_valid_t0)
    assert are_identical

    df["bit2_check"] = df.apply(lambda row: get_bit(row["aggr1"], 2), axis=1)
    are_identical = df["bit2_check"].equals(df.is_valid_tmax)
    assert are_identical


def test_build_hit_spms_basic(dsp_test_file_spm, tmp_dir):
    out_file = f"{tmp_dir}/L200-comm-20211130-phy-spms_hit.lh5"
    build_hit(
        dsp_test_file_spm,
        outfile=out_file,
        hit_config=f"{config_dir}/spms-hit-config.json",
        wo_mode="overwrite_file",
    )
    assert lh5.ls(out_file) == ["ch0", "ch1", "ch2"]
    assert lh5.ls(out_file, "ch0/") == ["ch0/hit"]
    assert lh5.ls(out_file, "ch0/hit/") == [
        "ch0/hit/energy_in_pe",
        "ch0/hit/quality_cut",
        "ch0/hit/trigger_pos",
    ]


def test_build_hit_spms_multiconfig(dsp_test_file_spm, tmp_dir):
    out_file = f"{tmp_dir}/L200-comm-20211130-phy-spms_hit.lh5"

    # append the tmp_dir to the start of paths in the spms-hit-multi-config.json
    with open(f"{config_dir}/spms-hit-multi-config.json") as f:
        configdict = json.load(f)
    for key in configdict.keys():
        configdict[key] = f"{config_dir}/" + configdict[key].split("/")[-1]
    newdict = json.dumps(configdict)
    with open(f"{tmp_dir}/spms-hit-multi-config.json", "w") as file:
        file.write(newdict)

    build_hit(
        dsp_test_file_spm,
        outfile=out_file,
        lh5_tables_config=f"{tmp_dir}/spms-hit-multi-config.json",
        wo_mode="overwrite",
    )
    assert lh5.ls(out_file) == ["ch0", "ch1", "ch2"]
    assert lh5.ls(out_file, "ch0/") == ["ch0/hit"]
    assert lh5.ls(out_file, "ch0/hit/") == [
        "ch0/hit/energy_in_pe",
        "ch0/hit/quality_cut",
        "ch0/hit/trigger_pos",
    ]


def test_build_hit_spms_calc(dsp_test_file_spm, tmp_dir):
    out_file = f"{tmp_dir}/L200-comm-20211130-phy-spms_hit.lh5"

    build_hit(
        dsp_test_file_spm,
        outfile=out_file,
        wo_mode="overwrite_file",
        lh5_tables_config=f"{config_dir}/spms-hit-a-config.json",
    )
    assert lh5.ls(out_file) == ["ch0", "ch1", "ch2"]
    assert lh5.ls(out_file, "ch0/") == ["ch0/hit"]
    assert lh5.ls(out_file, "ch0/hit/") == ["ch0/hit/energy_in_pe"]

    df0 = lh5.read_as("ch0/hit/energy_in_pe", out_file, "np")
    df1 = lh5.read_as("ch1/hit/energy_in_pe", out_file, "np")
    df2 = lh5.read_as("ch2/hit/energy_in_pe", out_file, "np")

    assert len(df0) == 5
    assert len(df1) == 5
    assert len(df2) == 5

    assert len(df0[0]) == 20
    assert len(df1[0]) == 20
    assert len(df2[0]) == 20

    assert np.nanmean(df0) == 0
    assert np.nanmean(df1) == 1
    assert np.nanmean(df2) == 2


def test_vov_input(lgnd_test_data, tmp_dir):
    infile = lgnd_test_data.get_path(
        "lh5/l200-p03-r000-phy-20230312T055349Z-tier_psp.lh5"
    )
    outfile = f"{tmp_dir}/LDQTA_r117_20200110T105115Z_cal_geds_hit.lh5"

    hit_config = {
        "outputs": ["a"],
        "operations": {
            "a": {
                "expression": "a + m * energies",
                "parameters": {"a": 0, "m": 1},
            }
        },
    }

    build_hit(
        infile, outfile=outfile, hit_config=hit_config, wo_mode="of", buffer_len=9999999
    )

    orig = lh5.read_as("ch1067205/dsp/energies", infile, "ak")
    data = lh5.read_as("ch1067205/hit/a", outfile, "ak")
    assert ak.all(data == orig)



"""
Uniform distributions for pygama
"""

import numba as nb
import numpy as np
from numba import prange

from pygama.math.functions.pygama_continuous import PygamaContinuous
from pygama.utils import numba_math_defaults as nb_defaults
from pygama.utils import numba_math_defaults_kwargs as nb_kwargs


@nb.njit(**nb_kwargs)
def nb_uniform_pdf(x: np.ndarray, a: float, b: float) -> np.ndarray:
    r"""
    Normalised uniform probability density function, w/ args: a, b. Its range of support is :math:`x\in[a,b]`. The pdf is computed as:


    .. math::
        pdf(x, a, b) = \begin{cases} \frac{1}{b-a} \quad , a\leq x\leq b \\ 0 \quad , \text{otherwise} \end{cases}


    As a Numba JIT function, it runs slightly faster than
    'out of the box' functions.

    Parameters
    ----------
    x
        The input data
    a
        The lower edge of the distribution
    b
        The upper edge of the distribution
    """
    b = a + b  # gives dist on [a, a+b] like scipy's does
    w = b - a
    p = np.empty_like(x, np.float64)
    for i in prange(x.shape[0]):
        if a <= x[i] <= b:
            p[i] = 1 / w
        else:
            p[i] = 0
    return p


@nb.njit(**nb_kwargs)
def nb_uniform_cdf(x: np.ndarray, a: float, b: float) -> np.ndarray:
    r"""
    Normalised uniform cumulative distribution, w/ args: a, b. Its range of support is :math:`x\in[a,b]`. The cdf is computed as:


    .. math::
        cdf(x, a, b) = \begin{cases} 0 \quad , x<a \\ \frac{x-a}{b-a} \quad , a\leq x\leq b  \\ 1 \quad , x>b \end{cases}


    As a Numba JIT function, it runs slightly faster than
    'out of the box' functions.

    Parameters
    ----------
    x
        The input data
    a
        The lower edge of the distribution
    b
        The upper edge of the distribution
    """

    b = a + b  # gives dist on [a, a+b] like scipy's does
    w = b - a
    p = np.empty_like(x, np.float64)
    for i in prange(x.shape[0]):
        if a <= x[i]:
            if x[i] <= b:
                p[i] = (x[i] - a) / w
            else:
                p[i] = 1
        else:
            p[i] = 0
    return p


@nb.njit(**nb_defaults(parallel=False))
def nb_uniform_scaled_pdf(x: np.ndarray, area: float, a: float, b: float) -> np.ndarray:
    r"""
    Scaled uniform probability distribution, w/ args: a, b.
    As a Numba JIT function, it runs slightly faster than
    'out of the box' functions.


    Parameters
    ----------
    x
        Input data
    area
        The number of counts in the signal
    a
        The lower edge of the distribution
    b
        The upper edge of the distribution
    """

    return area * nb_uniform_pdf(x, a, b)


@nb.njit(**nb_defaults(parallel=False))
def nb_uniform_scaled_cdf(x: np.ndarray, area: float, a: float, b: float) -> np.ndarray:
    r"""
    Uniform cdf scaled by the area, used for extended binned fits
    As a Numba JIT function, it runs slightly faster than
    'out of the box' functions.


    Parameters
    ----------
    x
        Input data
    area
        The number of counts in the signal
    a
        The lower edge of the distribution
    b
        The upper edge of the distribution
    """

    return area * nb_uniform_cdf(x, a, b)


class UniformGen(PygamaContinuous):

    def _get_support(self, a, b):
        return a, b

    def _argcheck(self, a, b):
        return b > a

    def __init__(self, *args, **kwargs):
        self.x_lo = None
        self.x_hi = None
        super().__init__(*args, **kwargs)

    def _pdf(self, x: np.ndarray, a, b) -> np.ndarray:
        return nb_uniform_pdf(x, a[0], b[0])

    def _cdf(self, x: np.ndarray, a, b) -> np.ndarray:
        return nb_uniform_cdf(x, a[0], b[0])

    def get_pdf(self, x: np.ndarray, a: float, b: float) -> np.ndarray:
        return nb_uniform_pdf(x, a, b)

    def get_cdf(self, x: np.ndarray, a: float, b: float) -> np.ndarray:
        return nb_uniform_cdf(x, a, b)

    def pdf_norm(
        self, x: np.ndarray, x_lo: float, x_hi: float, a: float, b: float
    ) -> np.ndarray:
        return self._pdf_norm(x, x_lo, x_hi, a, b)

    def cdf_norm(
        self, x: np.ndarray, x_lo: float, x_hi: float, a: float, b: float
    ) -> np.ndarray:
        return self._cdf_norm(x, x_lo, x_hi, a, b)

    def pdf_ext(
        self, x: np.ndarray, x_lo: float, x_hi: float, area: float, a: float, b: float
    ) -> np.ndarray:
        return np.diff(nb_uniform_scaled_cdf(np.array([x_lo, x_hi]), area, a, b))[
            0
        ], nb_uniform_scaled_pdf(x, area, a, b)

    def cdf_ext(self, x: np.ndarray, area: float, a: float, b: float) -> np.ndarray:
        return nb_uniform_scaled_cdf(x, area, a, b)

    def required_args(self) -> tuple[str, str]:
        return "a", "b"


uniform = UniformGen(a=0.0, b=1.0, name="uniform")



import numpy as np
from scipy.stats import norm as scipy_gaussian

from pygama.math.functions.gauss import gaussian
from pygama.math.functions.pygama_continuous import PygamaContinuous


def test_gaussian_pdf():
    x = np.arange(-10, 12)
    mu = 1
    sigma = 2

    y = gaussian.pdf(x, mu, sigma)
    y_direct = gaussian.get_pdf(x, mu, sigma)
    scipy_y = scipy_gaussian.pdf(x, mu, sigma)

    assert isinstance(gaussian, PygamaContinuous)
    assert np.allclose(y, scipy_y, rtol=1e-8)
    assert np.allclose(y_direct, scipy_y, rtol=1e-8)

    n_sig = 20
    x_lo = -100
    x_hi = 100

    y_sig, y_ext = gaussian.pdf_ext(x, x_lo, x_hi, n_sig, mu, sigma)
    assert np.allclose(y_ext, 20 * scipy_y, rtol=1e-8)
    assert np.allclose(y_sig, n_sig, rtol=1e-8)

    normalization = np.diff(scipy_gaussian.cdf(x[np.array([0, -1])], mu, sigma))
    y_norm = gaussian.pdf_norm(x, x[0], x[-1], mu, sigma)

    assert np.allclose(y_norm, scipy_y / normalization, rtol=1e-8)


def test_gaussian_cdf():
    x = np.arange(-10, 12)
    mu = 1
    sigma = 2

    y = gaussian.cdf(x, mu, sigma)
    y_direct = gaussian.get_cdf(x, mu, sigma)
    scipy_y = scipy_gaussian.cdf(x, mu, sigma)

    assert isinstance(gaussian, PygamaContinuous)
    assert np.allclose(y, scipy_y, rtol=1e-8)
    assert np.allclose(y_direct, scipy_y, rtol=1e-8)

    n_sig = 20
    y_ext = gaussian.cdf_ext(x, n_sig, mu, sigma)
    assert np.allclose(y_ext, 20 * scipy_y, rtol=1e-8)

    normalization = np.diff(scipy_gaussian.cdf(x[np.array([0, -1])], mu, sigma))
    y_norm = gaussian.cdf_norm(x, x[0], x[-1], mu, sigma)

    assert np.allclose(y_norm, scipy_y / normalization, rtol=1e-8)


def test_required_args():
    names = gaussian.required_args()
    assert names[0] == "mu"
    assert names[1] == "sigma"


def test_name():
    assert gaussian.name == "gaussian"
